import numpy as np
import cvxopt as cv
from cvxopt.modeling import op, variable, dot, matrix
import pdb
import json
import properscoring as ps



# autocorrelation function from
# https://stackoverflow.com/questions/643699/how-can-i-use-numpy-correlate-to-do-autocorrelation
def acf(x, length=20):
    return np.array([1]+[np.corrcoef(x[:-i], x[i:])[0,1]  \
        for i in range(1, length)])

# Mean Absolute Error calculator
def mae(obs,fc):
	val = np.mean(np.abs(fc-obs))
	return val 

# RMSE calculator
def rmse(obs,fc):
	val = np.sqrt(np.mean((fc-obs)**2.))
	return val 





# Generate a decision forecast (and evaluate it) for a reservoir management problem
# Here use stochastic optimization across ensemble (for deterministic use opttype=0 and ensemble of one)
def gen_decfc_ens(maxdraw,mindraw,maxcap,mincap,drawprice,shortprice,overtopprice,initres,obs,fc,opttype,summary):

	# Length of timeseries
	nt = len(obs)
	nens = len(fc[:,0])
		
	# Sanity check - do fc and obs match in length?
	if (nt != len(fc[0,:])):
		print('Forecast and obs are different lengths = ',nt, len(fc[0,:]))
		import pdb; pdb.set_trace() # stops the programme for debugging

	# Storage space for data
	rstore = np.zeros(nt)
	rdrawdown = np.zeros(nt)
	rshortfall = np.zeros(nt)
	rovertop = np.zeros(nt)

	# Step through each timestep of the forecasted period
	for ii in np.arange(0,nt,1):

		# Create storage space for the temporary 'solutions' from each ensemble member forecast (deterministic)
		# idfstore = np.zeros(nens)
		idfdrawdown = np.zeros(nens)
		idfshortfall = np.zeros(nens)
		idfovertop = np.zeros(nens)

		# Create storage space for the temporary 'evaluations' from each ensemble member forecast (scenarois)
		# idfstore = np.zeros(nens)
		ipfdrawdown = np.zeros(nens)
		ipfshortfall = np.zeros(nens)
		ipfovertop = np.zeros(nens)
		ipfnetincome = np.zeros(nens)
		ipfnetincomemax = np.zeros(nens)
		ipfnetincomemin = np.zeros(nens)
		ipfnetincomemean = np.zeros(nens)

		# Here using optimization over the ensemble.  I.e., take decisions based on
		# entire met ensemble.
		# First: solve the cost-optimization problem for each INDIVIDUAL ens member
		for jj in np.arange(0,nens,1):

			# Look forward to next timestep using the known store plus forecast inflow
    	# Decision variables are indicated using variable()
    	# Variables starting with 'i' are temporary locations to store values
			ifc = fc[jj,ii]
			if (ii == 0):
				istore = initres
			else:
				istore = rstore[ii-1]
			idrawdown = variable()
			ishortfall = variable()
			iovertop = variable()

    	# Constraints on the decision problem optimization (for this single ensemble member)
    	# c1, c2 = Drawdown must be within min-max limits
    	# c3, c4 = shortfall and overtop must be positive values
    	# c5 = Drawdown is limited by FORECAST input + available STORAGE
			# NB: code change for python3.8 - np.float() was deprecated in python3.8 for float()
			c1 = ( idrawdown <= maxdraw )
			c2 = ( idrawdown >= mindraw )
			c3 = ( ishortfall >= 0. )
			c4 = ( iovertop >= 0. )
			c5 = ( idrawdown <= float(istore+ifc-mincap) )
#			c5 = ( idrawdown <= np.float(istore+ifc-mincap) )

    	# Run the optimizer, minimizes the cost function subject to constraints
    	# Cost = - (revenue from drawdown minus losses from overflow and shortfall)
			# NB: code change for python 3.8 - np.float() was deprecated in python3.8 for float()
#			lp1 = op( (-idrawdown*np.float(drawprice)) + ishortfall*np.float(shortprice) + iovertop*np.float(overtopprice), [c1,c2,c3,c4,c5])
			lp1 = op( (-idrawdown*float(drawprice)) + ishortfall*float(shortprice) + iovertop*float(overtopprice), [c1,c2,c3,c4,c5])
			lp1.solve()

    	# Store the forecasted decision values (DETERMINISTIC)
			idfdrawdown[jj] = idrawdown.value[0]
			idfovertop[jj] = iovertop.value[0]
			idfshortfall[jj] = ishortfall.value[0]

    	# BUT! we don't know that this particular met scenario will occur.  So, calculate the netincome across each of the 
    	#	possible meteorological scenarios in the ensemble
			for kk in np.arange(0,nens,1):
			
				kfc = fc[kk,ii]
			
				ipfshortfall[kk] = np.abs(np.min([istore + kfc - idfdrawdown[jj], mincap]) - mincap)
				ipfovertop[kk] = np.max([istore + kfc - idfdrawdown[jj], maxcap]) - maxcap
				ipfdrawdown[kk] = idfdrawdown[jj] - ipfshortfall[kk]
				ipfnetincome[kk] = ipfdrawdown[kk]*drawprice - (ipfshortfall[kk]*shortprice + ipfovertop[kk]*overtopprice)

			# Now, calculate the best/worst/average net income that is realised given different met scenarios applied to this
			# decision
			ipfnetincomemax[jj] = np.max(ipfnetincome)
			ipfnetincomemin[jj] = np.min(ipfnetincome)
			ipfnetincomemean[jj] = np.mean(ipfnetincome)
			
			
		# Now apply the stochastic optimization criteria
		if ((opttype == 10) or (opttype == 20)):
		
			# Deterministic (10 = perfect, 20 = ensmean)
			idec = [True]
		
		elif (opttype == 30):
		
			# MAX(min income)
			idec = (ipfnetincomemin == np.max(ipfnetincomemin))

		elif (opttype == 40):

			# MAX(mean income)
			idec = (ipfnetincomemean == np.max(ipfnetincomemean))

		elif (opttype == 50):
		
			# MAX(max income)
			idec = (ipfnetincomemax == np.max(ipfnetincomemax))
	
			
		else:
		
			print('Optimization type not recognised',opttype)
			import pdb; pdb.set_trace() # stops the programme for debugging

		# Although not strictly essential, convert the idec boolean array to an integer index
		# if more than one equally good solution exists, pick just the first one
		# NB: code change for python3.8 -- make iind a simple float, not array in all cases
		iind = np.arange(0,len(idec))[idec]
#		if (len(iind) > 1):
		if (len(iind) >= 1):
			iind = iind[0]
		elif (len(iind) == 0):
			print('Error: No optimal solution found')
			import pdb; pdb.set_trace() # stops the programme for debugging


		# Resolve actual events for this timestep (nb: actual weather is different to all the cases in the ensemble)
    # Note: drawdown may be less than forecast if there is less input than expected
		# Note: store is updated by ACTUAL (observed) inflow minus the ACTUAL drawdown		
		rshortfall[ii] = np.abs(np.min([istore + obs[ii] - idfdrawdown[iind], mincap]) - mincap)
		rovertop[ii] = np.max([istore + obs[ii] - idfdrawdown[iind], maxcap]) - maxcap
		rdrawdown[ii] = idfdrawdown[iind] - rshortfall[ii]
		rstore[ii] = istore + obs[ii] - (rdrawdown[ii] + rovertop[ii])


	# Calculate the net income from the decisions taken and actual outcomes over whole timeseries
	rnetincome = rdrawdown*drawprice - (rshortfall*shortprice + rovertop*overtopprice)


	# Create datastructure to return to calling function
	# If the summary flag is set, then just return 'summary' statistics rather than
	# full timeseries
	if (summary):

		dec = {'nt':nt,
					 'optype':opttype,
		    	 'maxdraw':maxdraw,
        	 'mindraw':mindraw,
        	 'maxcap':maxcap,
        	 'mincap':mincap,
        	 'drawprice':drawprice,
        	 'shortprice':shortprice,
        	 'overtopprice':overtopprice,
        	 'initres':initres,
    	  	 'store_mean':np.mean(rstore),
    	  	 'store_std':np.std(rstore),
    	  	 'draw_mean':np.mean(rdrawdown),
    	  	 'short_mean':np.mean(rshortfall),
    	  	 'over_mean':np.mean(rovertop),
    	  	 'income_mean':np.mean(rnetincome),
					 'income_std':np.std(rnetincome)}
					 
	else:

		dec = {'nt':nt,
					 'optype':opttype,
		    	 'maxdraw':maxdraw,
        	 'mindraw':mindraw,
        	 'maxcap':maxcap,
        	 'mincap':mincap,
        	 'drawprice':drawprice,
        	 'shortprice':shortprice,
        	 'overtopprice':overtopprice,
        	 'initres':initres,
        	 'metobs':obs,
        	 'metfc':fc,
        	 'tt':np.arange(nt),
    	  	 'rstore':rstore,
    	  	 'rdrawdown':rdrawdown,
    	  	 'rshortfall':rshortfall,
    	  	 'rovertop':rovertop,
    	  	 'rnetincome':rnetincome}

	return dec





# Generate a forecast ensemble, based on Weigel 2009 model (following their
# notation for variable names).
def gen_metfc(nt, nens, alpha, beta, baserate, ar_pred, ar_unpred, ar_beta, seed):


	if (seed != 0):
		np.random.seed(seed) # Initialize the random number generator


	# Sanity check - ensure 'beta error' has zero autocorrelation for now
	# (not sure it makes physical sense to have non-zero beta error autocorr)
	if (np.abs(ar_beta) > 0.0000001):
		print('Beta autocorrelation non-zero, try with ar_beta = 0')
		import pdb; pdb.set_trace() # stops the programme for debugging


	# Sanity check - is beta in allowed range
	if (beta > np.sqrt(1.-alpha**2)):
		print('Max beta = ',np.sqrt(1.-alpha**2))
		import pdb; pdb.set_trace() # stops the programme for debugging


	# Ensemble spread
	ss_ens = np.sqrt(1.-alpha**2-beta**2)

	# create raw random number arrays (normal distributions; no AR)
	# scaling is such that, once used to produce the relevant qty,
	# the combined stddev should be unity	
	rand_pred = np.random.normal(scale=np.sqrt(1.-ar_pred**2.),size=nt)
	rand_unpred = np.random.normal(scale=np.sqrt(1.-ar_unpred**2.),size=nt)
	rand_beta = np.random.normal(scale=np.sqrt(1.-ar_beta**2.),size=nt)

	
	# Create the parts of the synthetic obs/fc that require memory / autoregress
	mu_xx = np.zeros(nt) # predictable part of obs
	ee_xx = np.zeros(nt) # unpredictable part of obs
	ee_beta = np.zeros(nt) # forecast error (offset beta)
	for ii in np.arange(1,nt,1):
		
		# Predictable signal (eq 8)...
		mu_xx[ii] = (mu_xx[ii-1] * ar_pred + rand_pred[ii])

		# Unpredictable signal (eq 9)... 
		ee_xx[ii] = (ee_xx[ii-1] * ar_unpred + rand_unpred[ii])

		# Forecast error/beta (eq 10)... 
		ee_beta[ii] = (ee_beta[ii-1] * ar_beta + rand_beta[ii])

	# Rescale to follow Weigel
	mu_xx = mu_xx * alpha
	ee_xx = ee_xx * np.sqrt(1.-alpha**2.)
	ee_beta = ee_beta * beta

	# Create forecast ensemble spread - NB: no memory/AR component
	ee_MM = np.random.normal(scale=ss_ens,size=(nens,nt)) 

	# Create forecast and obs pairs
	xx = mu_xx + ee_xx + baserate
	ff_xx = mu_xx + ee_beta + ee_MM + baserate

	# Forecast ensemble mean
	mu_ff = np.mean(ff_xx,0)

	# Calculate MAE, RMSE and CRPS
	ensmean_mae = mae(xx,mu_ff)
	ensmean_rmse = rmse(xx,mu_ff)
	ens_crps = ps.crps_ensemble(xx,np.transpose(ff_xx)).mean()


	# Create data structure to pass back to the calling function
	fc = {'nt':nt,
		  'nens':nens,
		  'baserate':baserate,
		  'alpha':alpha,
		  'beta':beta,
			'ar_pred':ar_pred,
			'ar_unpred':ar_unpred,
			'ar_beta':ar_beta,
      'tt':np.arange(0,nt,1),
		  'ss_ens':ss_ens,
			'rand_pred':rand_pred,
			'rand_unpred':rand_unpred,
			'rand_beta':rand_beta,
			'mae':ensmean_mae,
			'rmse':ensmean_rmse,
			'crps':ens_crps,
		  'mu_xx':mu_xx,
		  'ee_xx':ee_xx,
		  'xx':xx,
		  'ee_beta':ee_beta,
		  'ee_MM':ee_MM,
		  'ff_xx':ff_xx,
		  'mu_ff':mu_ff}

	return fc









# Generate a forecast ensemble, based on Weigel 2009 model (following their
# notation for variable names).
# Here, adding Len's underconfidence trick
def gen_metfc_uc(nt, nens, alpha, beta, baserate, ar_pred, ar_unpred, ar_beta, seed, uc):


	if (seed != 0):
		np.random.seed(seed) # Initialize the random number generator


	# Sanity check - ensure 'beta error' has zero autocorrelation for now
	# (not sure it makes physical sense to have non-zero beta error autocorr)
	if (np.abs(ar_beta) > 0.0000001):
		print('Beta autocorrelation non-zero, try with ar_beta = 0')
		import pdb; pdb.set_trace() # stops the programme for debugging


	# Sanity check - is beta in allowed range
	if (beta > np.sqrt(1.-alpha**2)):
		print('Max beta = ',np.sqrt(1.-alpha**2))
		import pdb; pdb.set_trace() # stops the programme for debugging


	# Ensemble spread
	ss_ens = np.sqrt(1.-alpha**2-beta**2)

	# create raw random number arrays (normal distributions; no AR)
	# scaling is such that, once used to produce the relevant qty,
	# the combined stddev should be unity	
	rand_pred = np.random.normal(scale=np.sqrt(1.-ar_pred**2.),size=nt)
	rand_unpred = np.random.normal(scale=np.sqrt(1.-ar_unpred**2.),size=nt)
	rand_beta = np.random.normal(scale=np.sqrt(1.-ar_beta**2.),size=nt)

	
	# Create the parts of the synthetic obs/fc that require memory / autoregress
	mu_xx = np.zeros(nt) # predictable part of obs
	ee_xx = np.zeros(nt) # unpredictable part of obs
	ee_beta = np.zeros(nt) # forecast error (offset beta)
	for ii in np.arange(1,nt,1):
		
		# Predictable signal (eq 8)...
		mu_xx[ii] = (mu_xx[ii-1] * ar_pred + rand_pred[ii])

		# Unpredictable signal (eq 9)... 
		ee_xx[ii] = (ee_xx[ii-1] * ar_unpred + rand_unpred[ii])

		# Forecast error/beta (eq 10)... 
		ee_beta[ii] = (ee_beta[ii-1] * ar_beta + rand_beta[ii])

	# Rescale to follow Weigel
	mu_xx = mu_xx * alpha
	ee_xx = ee_xx * np.sqrt(1.-alpha**2.)
	ee_beta = ee_beta * beta

	# Create forecast ensemble spread - NB: no memory/AR component
	ee_MM = np.random.normal(scale=ss_ens,size=(nens,nt)) 

	# Create forecast and obs pairs
	# Observation gets full part of predictable signal, forecast is 
	# reduced in variance
	xx = mu_xx + ee_xx + baserate
	
	
	#!!! HERE !!! DJB !!!
	import pdb; pdb.set_trace()
	print('!!! HERE !!! DJB !!!')
	ff_xx = (mu_xx * uc) + ((ee_beta + ee_MM) * np.sqrt(1. - uc**2.)) + baserate

	# Forecast ensemble mean
	mu_ff = np.mean(ff_xx,0)

	# Calculate MAE, RMSE and CRPS
	ensmean_mae = mae(xx,mu_ff)
	ensmean_rmse = rmse(xx,mu_ff)
	ens_crps = ps.crps_ensemble(xx,np.transpose(ff_xx)).mean()


	# Create data structure to pass back to the calling function
	fc = {'nt':nt,
		  'nens':nens,
		  'baserate':baserate,
		  'alpha':alpha,
		  'beta':beta,
			'ar_pred':ar_pred,
			'ar_unpred':ar_unpred,
			'ar_beta':ar_beta,
			'uc':uc,
      'tt':np.arange(0,nt,1),
		  'ss_ens':ss_ens,
			'rand_pred':rand_pred,
			'rand_unpred':rand_unpred,
			'rand_beta':rand_beta,
			'mae':ensmean_mae,
			'rmse':ensmean_rmse,
			'crps':ens_crps,
		  'mu_xx':mu_xx,
		  'ee_xx':ee_xx,
		  'xx':xx,
		  'ee_beta':ee_beta,
		  'ee_MM':ee_MM,
		  'ff_xx':ff_xx,
		  'mu_ff':mu_ff}

	return fc
