import numpy as np
import pdb
from batch_modelcore import *
from batch_io import *
from batch_s2s4e_data import *
import datetime
import properscoring as ps


# To run 'from the command line'
# module load anaconda
# source activate <your python3 environment>
# cd <relevant code directory>
# python3 <file>.py

# To run as a 'batch' job, see 
# https://research.reading.ac.uk/act/knowledgebase/racc-batch-jobs/
# module load anaconda
# source activate <your python3 environment>
# cd <relevant code directory>
# sbatch <script.sh> <== use this to submit the experiment 'job'
# squeue -u <userid> <== this to check it's running
# scancel <jobid>    <== use this to cancel the job if needed

# Experiment details
exptname = 'forShengxuan' # the 'name' of this experiment
exptdesc = 'A test experiment' # some notes to describe this experiment
code = 'forShengxuan.py' # name of this file
path='/storage/silver/metstudent/ug/bu804398/' # output will appear here
datadir = '/storage/silver/metstudent/ug/bu804398/s2s4e/country_daily_2m_temperature/' # your copy of the s2s4e data you want to use


# Select 'type' of decision-making to use
# Options: 
#  10: perfect foresight (deterministic)
#  20: ensemble mean forecast (deterministic)
#  30: max-min optimisation
#  40: max-mean optimisation
#  50: max-max optimisation 
# Note you can choose multiple, e.g., [10,20,30,40], model will loop over choices 
# with one optimisation method per output file
opttype = [10,20,30,40]

# Debug flag - set to True to see onscreen progress indicator, else False
debug = True

# Details about weather forecast input to use
# Here just using Nov-Mar forecasts from ECMWF, 2m temperature
season = ['11','12','01','02','03']
fileroot = 'ecmwf_s2s_coun_t2m_fcst_w_VI_and_rea_DAILY_stmon_'
s2s4e_checkvarnames(datadir+fileroot+'01.nc') # this prints the variable names in the file - you can use it to check the next line
var = 't2m' # weather variable - just the first bit, don't include _rea and _fcst


# Parameters to investigate, model will loop over your choices with all choices
# in a single output file

# Met forecast elements
country= ['Austria','United Kingdom'] # ['Austria', 'Belgium', 'Bulgaria', 'Croatia', 'Czech Republic', 'Denmark', 'Finland', 'France', 'Germany', 'Greece', 'Hungary', 'Ireland', 'Italy', 'Latvia', 'Lithuania', 'Luxembourg', 'Montenegro', 'Netherlands', 'Norway', 'Poland', 'Portugal', 'Romania', 'Slovakia', 'Slovenia', 'Spain', 'Sweden', 'Switzerland', 'United Kingdom']
leadtime = [0,7,14] # [0,7,14,21] how far ahead are we forecasting
windowlength = [7,14] # [0,7,14,21] what is the the length of the forecast window (e.g., weekly-average)
variability = [0.75,1.,1.25] # the forecast and obs are re-scaled to have a common stddev, this sets how variable it is


# Bucket model elements
maxcap=[0.00,4.00] # max capacity limit on reservoir before overtop	
drawprice=[0.5,1.0,2.0] # income gained per unit of drawdown


# Other decision model parameters
mindraw=0. # minimum drawdown of reservoir in one timestep
maxdraw=10. # maximum drawdown of reservoir in one timestep
mincap=0. # min capacity limit on reservoir before shortfall
shortprice=1. # cost per unit of shortfall
overtopprice=1. # cost per unit of overtopping
initres=0. # initial content of reservoir
baserate=5. # mean inflow into reservoir


#########################################################################
# !!! don't change code below here until you are confident to do so !!! #
#########################################################################

# Perform some initial set up actions 
stime = datetime.datetime.today()
cv.solvers.options['show_progress'] = False # Switch off the solver verbosity
mdi = -999.999 # missing data indicator



# Set up dimensions of arrays
d0 = len(country)
d0k = 'country'
d0v = country

d1 = len(leadtime)
d1k = 'leadtime'
d1v = leadtime

d2 = len(windowlength)
d2k = 'windowlength'
d2v = windowlength

d3 = len(variability)
d3k = 'variability'
d3v = variability

d4 = len(maxcap)
d4k = 'maxcap'
d4v = maxcap

d5 = len(drawprice)
d5k = 'drawprice'
d5v = drawprice


# Roll over the different optimization methods chosen
for oo in opttype:

	# Storage space.  Note only storing summary statistics not full timeseries
	met_mae = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	met_rmse = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	met_crps = np.ones([d0,d1,d2,d3,d4,d5])*mdi

	dec_income_mean = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_income_std = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_store_mean = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_store_std = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_over_mean = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_draw_mean = np.ones([d0,d1,d2,d3,d4,d5])*mdi
	dec_short_mean = np.ones([d0,d1,d2,d3,d4,d5])*mdi

	# Iterate over experiments
	# (This probably isn't "good" python but it works!)
	for jj in np.arange(0,len(leadtime),1):

		for kk in np.arange(0,len(windowlength),1):

			# Load in a set of fc for a given lead time / averaging window
			metfc, metobs, valid, launch, clist = s2s4e_getfc(datadir,fileroot,season,var,leadtime[jj],leadtime[jj]+windowlength[kk])

			# nt is the number of discrete forecast timesteps we have across years
			# start dates etc for a given lead time
			#
			# nens is the number of ensemble members in each forecast
			nt = metobs[0,:,:].size
			nens = metfc[0,0,0,:].size

			for ii in np.arange(0,len(country),1):

				for ll in np.arange(0,len(variability),1):

					# We want the forecast to 'resemble' the artificial data
					# usually used to drive the bucket model... so
					# rescale the variance of the metfc & observation as desired and
					# add a fixed base inflow so that forecast/obs are always positive
					icountry = clist.index(country[ii])
					sc_metfc = ((metfc[icountry,:,:,:] - np.mean(metfc[icountry,:,:,:])) /
												np.std(metfc[icountry,:,:,:] - np.mean(metfc[icountry,:,:,:]))) * variability[ll] + baserate

					sc_metobs = ((metobs[icountry,:,:] - np.mean(metobs[icountry,:,:])) /
												np.std(metobs[icountry,:,:] - np.mean(metobs[icountry,:,:]))) * variability[ll] + baserate

					# Ugly hack - prevent negative values
					sc_metobs.data[sc_metobs.data < 0.] = 0.0
					sc_metfc.data[sc_metfc.data < 0.] = 0.0

					# Calculate the ensemble mean forecast
					sc_metfc_ensmean = np.mean(sc_metfc,2)

					for mm in np.arange(0,len(maxcap),1):

						for nn in np.arange(0,len(drawprice),1):

							if (debug): print('Experiment: ',ii,' - ',jj,' - ',kk,' - ',ll, ' - ',mm,' - ',nn,' (',oo,')')


							# calculate	rmse etc for the forecasts					
							met_mae[ii,jj,kk,ll,mm,nn] = mae(np.resize(sc_metobs,(1,nt)),np.resize(sc_metfc_ensmean,(1,nt)))
							met_rmse[ii,jj,kk,ll,mm,nn] = rmse(np.resize(sc_metobs,(1,nt)),np.resize(sc_metfc_ensmean,(1,nt)))
							met_crps[ii,jj,kk,ll,mm,nn] = ps.crps_ensemble(np.resize(sc_metobs,nt),np.resize(sc_metfc,(nt,nens))).mean()

							# Run the decision-making algorithm using the perfect, ensmean (det), 
							# maxmin, maxmean, maxmax (stoch) methods as appropriate

							# Deterministic - perfect
							if (oo == 10):

								decfc = gen_decfc_ens(maxdraw,
													 						mindraw,
																			maxcap[mm],
																			mincap,
																			drawprice[nn],
																			shortprice,
																			overtopprice,
																			initres,
																			np.resize(sc_metobs,nt),
																			np.resize(sc_metobs,(1,nt)),
																			oo,
																			True)

							# Deterministic - ensmean
							if (oo == 20):

								decfc = gen_decfc_ens(maxdraw,
													 						mindraw,
																			maxcap[mm],
																			mincap,
																			drawprice[nn],
																			shortprice,
																			overtopprice,
																			initres,
																			np.resize(sc_metobs,nt),
																			np.resize(sc_metfc_ensmean,(1,nt)),
																			oo,
																			True)


							# Stoch - maxmin
							elif (oo == 30):

								decfc = gen_decfc_ens(maxdraw,
													 						mindraw,
																			maxcap[mm],
																			mincap,
																			drawprice[nn],
																			shortprice,
																			overtopprice,
																			initres,
																			np.resize(sc_metobs,nt),
																			np.transpose(np.resize(sc_metfc,(nt,nens))),
																			oo,
																			True)


							# Stoch - maxmean
							elif (oo == 40):

								decfc = gen_decfc_ens(maxdraw,
													 						mindraw,
																			maxcap[mm],
																			mincap,
																			drawprice[nn],
																			shortprice,
																			overtopprice,
																			initres,
																			np.resize(sc_metobs,nt),
																			np.transpose(np.resize(sc_metfc,(nt,nens))),
																			oo,
																			True)


							# Stoch - maxmax
							elif (oo == 50):

								decfc = gen_decfc_ens(maxdraw,
													 						mindraw,
																			maxcap[mm],
																			mincap,
																			drawprice[nn],
																			shortprice,
																			overtopprice,
																			initres,
																			np.resize(sc_metobs,nt),
																			np.transpose(np.resize(sc_metfc,(nt,nens))),
																			oo,
																			True)

							# In each case calculate then store a copy of the summary variables
							dec_income_mean[ii,jj,kk,ll,mm,nn] = decfc['income_mean']
							dec_income_std[ii,jj,kk,ll,mm,nn] = decfc['income_std']
							dec_store_mean[ii,jj,kk,ll,mm,nn] = decfc['store_mean']
							dec_store_std[ii,jj,kk,ll,mm,nn] = decfc['store_std']
							dec_over_mean[ii,jj,kk,ll,mm,nn] = decfc['over_mean']
							dec_draw_mean[ii,jj,kk,ll,mm,nn] = decfc['draw_mean']
							dec_short_mean[ii,jj,kk,ll,mm,nn] = decfc['short_mean']



	# Save the data out
	filename=path+'/'+exptname+'_opttype'+str(oo)+'.json'
	data = {'exptname':exptname,
					'desc':exptdesc,
					'filename':filename,
					'path':path,
					'code':code,
					'dataroot':datadir+fileroot,
					'mdi':mdi,
					'var':var,
					'season':np.asarray(season).tolist(),
					'nt':np.asarray(nt).tolist(),
					'nens':np.asarray(nens).tolist(),
					'baserate':np.asarray(baserate).tolist(),
					'variability':np.asarray(variability).tolist(),
					'country':np.asarray(country).tolist(),
					'leadtime':np.asarray(leadtime).tolist(),
					'windowlength':np.asarray(windowlength).tolist(),
					'opttype':np.asarray(oo).tolist(),
					'maxdraw':np.asarray(maxdraw).tolist(),
					'mindraw':np.asarray(mindraw).tolist(),
					'maxcap':np.asarray(maxcap).tolist(),
					'mincap':np.asarray(mincap).tolist(),
					'drawprice':np.asarray(drawprice).tolist(),
					'shortprice':np.asarray(shortprice).tolist(),
					'overtopprice':np.asarray(overtopprice).tolist(),
					'initres':np.asarray(initres).tolist(),
					'd0':d0,
					'd0k':d0k,
					'd0v':np.asarray(d0v).tolist(),
					'd1':d1,
					'd1k':d1k,
					'd1v':np.asarray(d1v).tolist(),
					'd2':d2,
					'd2k':d2k,
					'd2v':np.asarray(d2v).tolist(),
					'd3':d3,
					'd3k':d3k,
					'd3v':np.asarray(d3v).tolist(),
					'd4':d4,
					'd4k':d4k,
					'd4v':np.asarray(d4v).tolist(),
					'd5':d5,
					'd5k':d5k,
					'd5v':np.asarray(d5v).tolist(),
					'mae':np.asarray(met_mae).tolist(),
					'rmse':np.asarray(met_rmse).tolist(),
					'crps':np.asarray(met_crps).tolist(),
					'income_mean':np.asarray(dec_income_mean).tolist(),
					'income_std':np.asarray(dec_income_std).tolist(),
					'store_mean':np.asarray(dec_store_mean).tolist(),
					'store_std':np.asarray(dec_store_std).tolist(),
					'over_mean':np.asarray(dec_over_mean).tolist(),
					'draw_mean':np.asarray(dec_draw_mean).tolist(),
					'short_mean':np.asarray(dec_short_mean).tolist()}
	saveexpt(filename,data)




etime = datetime.datetime.today()
print('Start:',stime)
print('End:',etime)



