import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.optimize as optimize
import scipy.stats as stats
from random import *
from collections import Counter
import datetime
import cvxopt as cv
from cvxopt.modeling import op, variable, dot, matrix
from netCDF4 import Dataset as ncfile
import pdb
from batch_modelcore import *
from batch_io import *


# Path to where the plots will appear
path='../figs/'

# Meteorological forecast parameters
nt = 100 # number of timesteps in forecast period
nens = 15 # number of ensemble members in meteorological forecast
alpha = 0.66 # Weigel's alpha
beta = 0.0 # Weigel's beta
arp = 0.05 # AR1 (autocorrelation/persistence) of predictable component
aru = 0.05 # AR1 (autocorrelation/persistence) of unpredictable component

# Decision model parameters
maxcap=4. # max capacity limit on reservoir before overtop
drawprice=0.5 # income gained per unit of drawdown

# Random number seed (selects the 'weather' in the Weigel model)
seed = 1


# Other standard model parameters
baserate = 5. # background value to 'add' to forecast (mean state)
arb = 0. # AR1 (autocorrelation/persistence) of Weigel's beta term, don't change!
maxdraw=10. # maximum drawdown of reservoir in one timestep
mindraw=0. # minimum drawdown of reservoir in one timestep
mincap=0. # min capacity limit on reservoir before shortfall
shortprice=1. # cost per unit of shortfall
overtopprice=1. # cost per unit of overtopping
initres=0. # initial content of reservior

# Perform some initial set up actions (don't change this bit!)
cv.solvers.options['show_progress'] = False # Switch off the solver verbosity

# Call our forecast generator function (based on Weigel et al. 2009, sec3)
# Note this returns a data structure.
# You can see how this is built/laid out by looking at the end of the
# gen_metfc() function call.
metfc = gen_metfc(nt,nens,alpha,beta,baserate,arp,aru,arb,seed)

# Run the decision-making algorithm using the forecast
# This is where the really interesting bit starts!

# First, let's run the decision model with a 'perfect' forecast
# i.e., we manage the reservoir assuming we know the next timestep's weather
# This is the 'best' management we can achieve
decfc_perfect = gen_decfc_ens(maxdraw,
															mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															metfc['xx'],
															np.resize(metfc['xx'],(1,nt)),
															10,
															False)

# Now, let's run the decision model with an 'imperfect' forecast
# first, ensemble mean forecast
decfc_ensmean = gen_decfc_ens(maxdraw,
															mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															metfc['xx'],
															np.resize(metfc['mu_ff'],(1,nt)),
															20,
															False)


# second, the max-mean forecast
decfc_maxmin = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															metfc['xx'],
															metfc['ff_xx'],
															30,
															False)


# third, the max-mean forecast
decfc_maxmean = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															metfc['xx'],
															metfc['ff_xx'],
															40,
															False)


# finally, the max-max forecast
decfc_maxmax = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															metfc['xx'],
															metfc['ff_xx'],
															50,
															False)


# Let's now do some simple plots to illustrate what we've created


# Time series of the meteorological forecast
fig = plt.figure(figsize=(6, 12))

ax1 = fig.add_subplot(2,1,1)
xvals = metfc['tt']
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Timestep', fontsize='x-large')
ax1.set_ylabel('Metvar', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,metfc['ff_xx'][0,:],label='metfc ens member',color="red",linewidth=1.)
for ii in np.arange(1,nens): ax1.plot(xvals,metfc['ff_xx'][ii,:],color="red",linewidth=1.)
ax1.plot(xvals,metfc['mu_ff'],color="purple",label='metfc ens mean', linewidth=3.)
ax1.plot(xvals,metfc['xx'],color="black",label='met obs')
ax1.legend(loc='upper right', fontsize='medium')

ax2 = fig.add_subplot(2,1,2)
xvals = metfc['tt']
minx = np.min(xvals)
maxx = np.max(xvals)
stepx = 100
miny = 0.
maxy = 15.01
stepy = 5.
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Timestep', fontsize='x-large')
ax2.set_ylabel('Metvar', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,metfc['mu_ff'],color="purple",label='metfc ens mean', linewidth=3.)
ax2.plot(xvals,metfc['xx'],color="black",label='met obs')
ax2.legend(loc='upper right', fontsize='medium')

plt.savefig(path+'singleFL_met_timeseries.png')
plt.close(fig)


# PDFs of the ensemble forecast at timesteps 0, 1, 2
#
# Note: because we know how the 'truth' is constructed (Weigel's eqn 9)
#       we can also estimate the 'true' unpredictable noise around the predictable signal
#       hence can construct 'observational' KDEs
# NB: code change for python3.8 -- scipy.stats.kde.gaussian_kde => scipy.stats.gaussian_kde
met_kde_space = np.linspace( 0, 15, 31 )
met_singlefc0_kde = scipy.stats.gaussian_kde(metfc['ff_xx'][:,0]) (met_kde_space)
met_singlefc1_kde = scipy.stats.gaussian_kde(metfc['ff_xx'][:,1]) (met_kde_space)
met_singlefc2_kde = scipy.stats.gaussian_kde(metfc['ff_xx'][:,2]) (met_kde_space)
#met_singlefc0_kde = scipy.stats.kde.gaussian_kde(metfc['ff_xx'][:,0]) (met_kde_space)
#met_singlefc1_kde = scipy.stats.kde.gaussian_kde(metfc['ff_xx'][:,1]) (met_kde_space)
#met_singlefc2_kde = scipy.stats.kde.gaussian_kde(metfc['ff_xx'][:,2]) (met_kde_space)

true_noise = np.random.normal(scale=np.sqrt(1.-metfc['alpha']**2.),size=1000) # same as ee_xx
met_true0_kde = scipy.stats.gaussian_kde(metfc['mu_xx'][0]+true_noise+metfc['baserate']) (met_kde_space)
met_true1_kde = scipy.stats.gaussian_kde(metfc['mu_xx'][1]+true_noise+metfc['baserate']) (met_kde_space)
met_true2_kde = scipy.stats.gaussian_kde(metfc['mu_xx'][2]+true_noise+metfc['baserate']) (met_kde_space)
#met_true0_kde = scipy.stats.kde.gaussian_kde(metfc['mu_xx'][0]+true_noise+metfc['baserate']) (met_kde_space)
#met_true1_kde = scipy.stats.kde.gaussian_kde(metfc['mu_xx'][1]+true_noise+metfc['baserate']) (met_kde_space)
#met_true2_kde = scipy.stats.kde.gaussian_kde(metfc['mu_xx'][2]+true_noise+metfc['baserate']) (met_kde_space)

fig = plt.figure(figsize=(18, 6))

ax1 = fig.add_subplot(1,3,1)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Metvar', fontsize='x-large')
ax1.set_ylabel('KDE', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,met_singlefc0_kde,color="red",label='metfc ens (time = 0)', linewidth=1.)
ax1.plot(xvals,met_true0_kde,color="black",label='truth (time = 0)', linewidth=1.)
ax1.plot([metfc['mu_ff'][0],metfc['mu_ff'][0]],[miny,maxy],color="red",label='met ens mean (time = 0)', linewidth=3.)
ax1.plot([metfc['mu_xx'][0]+metfc['baserate'],metfc['mu_xx'][0]+metfc['baserate']],[miny,maxy],color="black",label='met obs pred (time = 0)', linewidth=3.)
ax1.plot([metfc['xx'][0],metfc['xx'][0]],[miny,maxy],color="blue",label='met obs (time = 0)', linewidth=3.)
ax1.legend(loc='upper right', fontsize='medium')

ax2 = fig.add_subplot(1,3,2)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Metvar', fontsize='x-large')
ax2.set_ylabel('KDE', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,met_singlefc1_kde,color="red",label='metfc ens (time = 1)', linewidth=1.)
ax2.plot(xvals,met_true1_kde,color="black",label='truth (time = 1)', linewidth=1.)
ax2.plot([metfc['mu_ff'][1],metfc['mu_ff'][1]],[miny,maxy],color="red",label='met ens mean (time = 1)', linewidth=3.)
ax2.plot([metfc['mu_xx'][1]+metfc['baserate'],metfc['mu_xx'][1]+metfc['baserate']],[miny,maxy],color="black",label='met obs pred (time = 1)', linewidth=3.)
ax2.plot([metfc['xx'][1],metfc['xx'][1]],[miny,maxy],color="blue",label='met obs (time = 1)', linewidth=3.)
ax2.legend(loc='upper right', fontsize='medium')

ax3 = fig.add_subplot(1,3,3)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax3.axis([minx,maxx,miny,maxy])
ax3.tick_params(direction='out', which='both', labelsize='x-large')
ax3.set_xlabel('Metvar', fontsize='x-large')
ax3.set_ylabel('KDE', fontsize='x-large')
ax3.set_xticks(np.arange(minx,maxx,stepx))
ax3.set_yticks(np.arange(miny,maxy,stepy))
ax3.plot(xvals,met_singlefc2_kde,color="red",label='metfc ens (time = 2)', linewidth=1.)
ax3.plot(xvals,met_true2_kde,color="black",label='truth (time = 2)', linewidth=1.)
ax3.plot([metfc['mu_ff'][2],metfc['mu_ff'][2]],[miny,maxy],color="red",label='met ens mean (time = 2)', linewidth=3.)
ax3.plot([metfc['mu_xx'][2]+metfc['baserate'],metfc['mu_xx'][2]+metfc['baserate']],[miny,maxy],color="black",label='met obs pred (time = 2)', linewidth=3.)
ax3.plot([metfc['xx'][2],metfc['xx'][2]],[miny,maxy],color="blue",label='met obs (time = 2)', linewidth=3.)
ax3.legend(loc='upper right', fontsize='medium')

plt.savefig(path+'singleFL_met_pdfs.png')
plt.close(fig)




# Time series of the decision model
fig = plt.figure(figsize=(12, 12))

ax1 = fig.add_subplot(2,2,1)
xvals = metfc['tt']
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Timestep', fontsize='x-large')
ax1.set_ylabel('Decvar', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_perfect['rnetincome'],color="black",label='realised net income (perfect fc)')
ax1.plot(xvals,decfc_perfect['rdrawdown'],color="orange",label='realised drawdown')
ax1.plot(xvals,decfc_perfect['rstore'],color="blue",label='realised storage')
ax1.plot(xvals,decfc_perfect['rovertop'],color="green",label='realised overtopping')
ax1.plot(xvals,decfc_perfect['rshortfall'],color="red",label='realised shortfall')
ax1.legend(loc='upper right', fontsize='medium')


ax2 = fig.add_subplot(2,2,2)
xvals = metfc['tt']
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Timestep', fontsize='x-large')
ax2.set_ylabel('Decvar', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,decfc_ensmean['rnetincome'],color="black",label='realised net income (ensmean fc)')
ax2.plot(xvals,decfc_ensmean['rdrawdown'],color="orange",label='realised drawdown')
ax2.plot(xvals,decfc_ensmean['rstore'],color="blue",label='realised storage')
ax2.plot(xvals,decfc_ensmean['rovertop'],color="green",label='realised overtopping')
ax2.plot(xvals,decfc_ensmean['rshortfall'],color="red",label='realised shortfall')
ax2.legend(loc='upper right', fontsize='medium')

ax3 = fig.add_subplot(2,2,3)
xvals = metfc['tt']
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax3.axis([minx,maxx,miny,maxy])
ax3.tick_params(direction='out', which='both', labelsize='x-large')
ax3.set_xlabel('Timestep', fontsize='x-large')
ax3.set_ylabel('Decvar', fontsize='x-large')
ax3.set_xticks(np.arange(minx,maxx,stepx))
ax3.set_yticks(np.arange(miny,maxy,stepy))
ax3.plot(xvals,decfc_maxmin['rnetincome'],color="black",label='realised net income (stoch maxmin)')
ax3.plot(xvals,decfc_maxmin['rdrawdown'],color="orange",label='realised drawdown')
ax3.plot(xvals,decfc_maxmin['rstore'],color="blue",label='realised storage')
ax3.plot(xvals,decfc_maxmin['rovertop'],color="green",label='realised overtopping')
ax3.plot(xvals,decfc_maxmin['rshortfall'],color="red",label='realised shortfall')
ax3.legend(loc='upper right', fontsize='medium')

ax4 = fig.add_subplot(2,2,4)
xvals = metfc['tt']
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax4.axis([minx,maxx,miny,maxy])
ax4.tick_params(direction='out', which='both', labelsize='x-large')
ax4.set_xlabel('Timestep', fontsize='x-large')
ax4.set_ylabel('Decvar', fontsize='x-large')
ax4.set_xticks(np.arange(minx,maxx,stepx))
ax4.set_yticks(np.arange(miny,maxy,stepy))
ax4.plot(xvals,decfc_maxmean['rnetincome'],color="black",label='realised net income (stoch maxmean)')
ax4.plot(xvals,decfc_maxmean['rdrawdown'],color="orange",label='realised drawdown')
ax4.plot(xvals,decfc_maxmean['rstore'],color="blue",label='realised storage')
ax4.plot(xvals,decfc_maxmean['rovertop'],color="green",label='realised overtopping')
ax4.plot(xvals,decfc_maxmean['rshortfall'],color="red",label='realised shortfall')
ax4.legend(loc='upper right', fontsize='medium')


plt.savefig(path+'singleFL_dec_timeseries.png')
plt.close(fig)



# Scatter plot of income vs metinflow
fig = plt.figure(figsize=(12, 12))

ax1 = fig.add_subplot(2,2,1)
xvals = metfc['xx']
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_perfect['rnetincome'],'xk',label='(perfect fc)')
ax1.legend(loc='upper right', fontsize='medium')

ax1 = fig.add_subplot(2,2,2)
xvals = metfc['xx']
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_ensmean['rnetincome'],'xk',label='(ensmean fc)')
ax1.legend(loc='upper right', fontsize='medium')

ax1 = fig.add_subplot(2,2,3)
xvals = metfc['xx']
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_maxmin['rnetincome'],'xk',label='(maxmin)')
ax1.legend(loc='upper right', fontsize='medium')


ax1 = fig.add_subplot(2,2,4)
xvals = metfc['xx']
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_maxmean['rnetincome'],'xk',label='(maxmean)')
ax1.legend(loc='upper right', fontsize='medium')



plt.savefig(path+'singleFL_scatter.png')
plt.close(fig)
